from .objects import Table, Field
from sqlite3 import dbapi2 as sqlite
from pathlib import Path


def dict_factory(cursor, row):
    return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}


class DB:
    conn = None
    cursor = None
    driver = None
    tables = None
    url = None

    def __init__(self, db_path: str):
        driver, db_path = db_path.split(":/")

        if DB.driver is None:
            DB.driver = self._get_driver(driver)

        if DB.conn is None and DB.cursor is None:
            db_path = Path(db_path)

            # make sure the folders to the db and the db itself exists
            db_folder_path = Path(str(db_path)[: -len(str(db_path.name))])
            db_folder_path.mkdir(parents=True, exist_ok=True)
            db_path.touch(exist_ok=True)

            DB.conn = sqlite.connect(db_path)
            DB.conn.row_factory = dict_factory

            DB.conn = DB.conn
            DB.cursor = DB.conn.cursor()
            DB.tables = {}
            DB.url = str(db_path)

    def create_tables(self):
        id_field = Field(field_type="INTEGER PRIMARY KEY AUTOINCREMENT")

        for table in Table.__subclasses__():
            table_name = (
                table.table_name if "table_name" in vars(table) else table.__name__
            )

            table_fields = {
                var: vars(table)[var]
                for var in vars(table)
                if isinstance(vars(table)[var], Field)
            }

            table_fields["id"] = id_field

            for field_name, field in table_fields.items():
                self._set_field(field, field_name, table_name)

            tbl = Table(
                db=self,
                conn=DB.conn,
                cursor=DB.cursor,
                driver=DB.driver,
                table_name=table_name,
                fields=[value for value in table_fields.values()],
            )

            DB.tables[table_name] = tbl
            self._create_table(table_name, *[field for field in table_fields.values()])

    def _set_field(self, field, field_name, table_name):
        field.field_name = field_name
        field.table_name = table_name
        field.cursor = DB.cursor
        field.conn = DB.conn
        field.driver = DB.driver
        field.db = self

    @classmethod
    def _get_driver(cls, driver: str):
        match driver:
            case "sqlite":
                from .drivers.sqlite import SqliteDriver
                from .dialects.sqlite import SqliteDialect

                return SqliteDriver(SqliteDialect)
        raise Exception(f"{driver} is not a valid driver!")

    @classmethod
    def _create_table(cls, table_name: str, *fields: Field):
        sql = f"CREATE TABLE IF NOT EXISTS {table_name} ("

        for field in fields:
            sql += f"{field.field_name} {field.field_type}, "
        sql = sql[:-2] + ")"

        DB.cursor.execute(sql)

    def __getattribute__(self, item):
        try:
            return super(DB, self).__getattribute__(item)
        except Exception as e:
            if item in DB.tables.keys():
                return DB.tables[item]
            raise e



class Query:
    def __init__(
        self,
        *,
        db=None,
        conn=None,
        cursor=None,
        dialect=None,
        driver=None,
        operator="=",
        first=None,
        second=None,
        tbl_name: str = None,
    ):
        self.db = db
        self.conn = conn
        self.cursor = cursor
        self.dialect = dialect
        self.driver = driver
        self.operator = operator
        self.first = first
        self.second = second
        self.table_name = tbl_name

    def __tables__(self):
        unpacked_query = self.dialect.unpack(self)
        return self.driver._unpacked_as_sql(unpacked_query)["tables"]

    def insert(self, **kwargs) -> None:
        """
        NOTE ALL FIELDS ARE REQUIRED TO USE THE INSERT CURRENTLY
        :param kwargs: list of values that will be inserted into the table
        :return:
        """
        sql = self.driver.insert(table_name=self.table_name, items=kwargs)

        self.cursor.execute(sql, tuple(kwargs.values()))
        self.conn.commit()

    def select(self, *fields: tuple, distinct=False, orderby=None) -> list[dict]:
        """
        Selects fields from the database and returns the result as a list of dicts
        :param fields: list of fields you want to select from the database
        :param distinct: use if you only want 1 of any value
        :param orderby: orders the result by the specified field
        :return:
        """

        # generate the select statement from what we have generated above
        sql = self.driver.select_sql(
            *fields,
            table_name=self.table_name,
            query=self,
            distinct=distinct,
            orderby=orderby,
        )

        return self.cursor.execute(sql).fetchall()

    def update(self, **kwargs):
        sql = self.driver.update(query=self, **kwargs)

        self.cursor.execute(sql, tuple(kwargs.values()))
        self.conn.commit()

    def delete(self):
        sql = self.driver.delete(query=self)

        self.cursor.execute(sql)
        self.conn.commit()

    def __and__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=self.dialect,
            operator=self.dialect.and_,
            driver=self.driver,
            first=self,
            second=other,
        )

    def __or__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=self.dialect,
            operator=self.dialect.or_,
            driver=self.driver,
            first=self,
            second=other,
        )



from .sql import SQLDialect
from ..query import Query


class SqliteDialect(SQLDialect):
    @classmethod
    def unpack(cls, query: Query) -> dict:
        unpacked_query = {'fields': [], 'stmts': []}

        if isinstance(query.first, Query):
            unpacked = cls.unpack(query.first)
            unpacked_query['fields'] += unpacked['fields']
            unpacked_query['stmts'] += unpacked['stmts']
        else:
            unpacked_query['fields'].append(query.first)

        unpacked_query['stmts'].append(query.operator)

        if isinstance(query.second, Query):
            unpacked = cls.unpack(query.second)
            unpacked_query['fields'] += unpacked['fields']
            unpacked_query['stmts'] += unpacked['stmts']
        else:
            unpacked_query['fields'].append(query.second)

        return unpacked_query



class SqliteDriver:
    def __init__(self, dialect):
        self.dialect = dialect

        # removes the amount of table from the select bc they are already selected
        self.tables_selected = 0

    def update(self, query, **kwargs):
        if len(kwargs) == 0:
            raise Exception("Need to give up args to update!")

        unpacked = self.dialect.unpack(query)

        # remove first table from select
        self.tables_selected = 1

        # make the select 1 due to speedup performances
        tables, where = self._unpacked_as_sql(unpacked).values()

        if len(tables) > 1:
            raise RuntimeError("Cannot select more then 1 table in a update statement!")

        # The join statement will start with the first table, the order of these tables should not matter
        # So we just select the first table and join the rest of the tables
        update_vals = ",".join([f"{key}=?" for key in kwargs.keys()])

        return self._update(first_table=tables[0], update_vals=update_vals, where=where)

    def delete(self, query):
        unpacked = self.dialect.unpack(query)

        # remove first table from select
        self.tables_selected = 1

        # fields needs to be 1
        tables, where = self._unpacked_as_sql(unpacked).values()

        if len(tables) > 1:
            raise RuntimeError("Cannot select more then 1 table in a update statement!")

        # generate sql based on prev calculated vals
        return self._delete(first_table=tables[0], select=where)

    def select_sql(
        self,
        *fields: tuple | int,
        query=None,
        table_name=None,
        distinct=False,
        orderby=None,
    ):
        # if no fields are specified, select all fields
        fields = [str(field) for field in fields] if len(fields) != 0 else "*"

        # unpack the query into a dict, note that the query uses a tree structure.
        # So a query can have a query as a child, which can have a query as a child, etc.
        # This unpacks the tree structure into a list of fields and a list of statements
        unpacked = self.dialect.unpack(query)
        if unpacked["fields"][0] is None:
            tables, where = self._unpacked_as_sql(unpacked).values()
            # where not doing a where stmt here!
            where = ""
        else:
            tables, where = self._unpacked_as_sql(unpacked).values()

        # if no table is found by unpacking the query, use the table_name specified in the select method
        # this is due to the user not using a where query and only db.table.select()
        tables = tables[self.tables_selected : :]

        # the join statement will start with the first table, the order of these tables should not matter
        # so we just select the first table and join the rest of the tables
        if len(tables) != 0:
            join = f"{tables[0]} "
        else:
            join = f"{table_name} {table_name}"
            where = where.replace(table_name, f"{table_name}")

        # however we can only join the tables if there are more then 1 table
        if len(tables) > 1:
            join += self._tables_to_join(tables)

        # if there is a where statement, add it to the join statement
        fields = ", ".join(fields) if isinstance(fields, list) else fields
        distinct = "DISTINCT " if distinct else ""
        orderby = f"ORDER BY {orderby}" if orderby else ""

        return self._select(
            fields=fields, tables=join, where=where, distinct=distinct, orderby=orderby
        )

    @classmethod
    def _tables_to_join(cls, tables):
        """
        make it so that every table is joined
        """
        return f"".join([f" INNER JOIN {table}" for table in tables[1:]])

    def _unpacked_as_sql(self, unpacked: dict) -> dict:
        sql = {"tables": [], "where": " WHERE "}

        while unpacked["fields"]:
            # grab 2 fields and a stmt to make an expr in the future
            field1 = str(unpacked["fields"].pop(0))
            field2 = str(unpacked["fields"].pop(0))
            stmt = unpacked["stmts"].pop(0)

            field_sql_dict = self.field_to_sql_dict(field1, sql["tables"])
            sql["where"] += field_sql_dict["field"]

            if field_sql_dict["table"] is not None:
                sql["tables"].append(field_sql_dict["table"])

            field_sql_dict = self.field_to_sql_dict(field2, sql["tables"])
            sql["where"] += f"{stmt} {field_sql_dict['field']}"

            if field_sql_dict["table"] is not None:
                sql["tables"].append(field_sql_dict["table"])

            if unpacked["stmts"]:
                sql["where"] += unpacked["stmts"].pop(0) + " "

        return sql

    @classmethod
    def field_to_sql_dict(cls, field: str, sql_tables: list):
        translated_field = cls._translate_field(field)
        table = (
            translated_field["table"]
            if translated_field["table"] and translated_field["table"] not in sql_tables
            else None
        )

        return dict(field=translated_field["field"], table=table)

    @classmethod
    def _translate_field(cls, field):
        if "." in str(field):
            table = field.split(".")[0]
            field = f"{field} "
        elif str(field).replace("-", "").isdigit():
            table = None
            field = f"{field} "
        else:
            table = None
            field = f"'{field}' "

        return dict(table=table, field=field)

    @classmethod
    def _set(cls, values: dict[str, str | bytes]):
        """
        turns a set of values into a key = value statement.
        example -> {"a": 2} returns "a=2"
        example2 -> {"a": 2, "b": "hi"} returns "a=2, b='hi'"
        """
        sql = ""
        first_done = False
        for key, value in values.items():
            if key == "id":
                continue

            if first_done:
                sql += ", "
            if str(value).isdigit():
                sql += f"{key} = {value} "
            else:
                sql += f"{key} = '{value}'"

            first_done = True
        return sql

    def insert(self, table_name: str, items: dict):
        return f"INSERT INTO {table_name} ({','.join(items.keys())}) VALUES ({','.join(['?' for _ in items.keys()])})"

    def _select(self, fields, tables, where=None, distinct=None, orderby=None):
        return f"SELECT {distinct}{fields} FROM {tables}{where}{orderby}"

    def _update(self, first_table: str, update_vals: str, where: str):
        return f"UPDATE {first_table} SET {update_vals} {where}"

    def _delete(self, first_table: str, select: str):
        return f"DELETE FROM {first_table} {select}"



class SQLDialect:
    equals = '='
    and_ = 'AND'
    or_ = 'OR'
    neq = '!='
    lt = '<'
    le = '<='
    gt = '>'
    ge = '>='

    @classmethod
    def sql_str(cls, value):
        return f"'{value}'"



from .dialects.sqlite import SqliteDialect
from .query import Query


class Field:
    def __init__(self, field_type):
        self.db = None
        self.conn = None
        self.cursor = None
        self.driver = None
        self.table_name = ""
        self.field_name = ""
        self.field_type = self._translate_type(field_type)

    def _translate_type(self, field_type):
        field_type_mapping = {
            "int": "INTEGER",
            "string": "VARCHAR(255)",
            "float": "FLOAT",
            "bool": "BOOLEAN",
            "date": "DATE",
            "datetime": "DATETIME",
            "time": "TIME",
            "image": "IMAGE",
        }
        sql_type = field_type_mapping.get(field_type.lower(), None)
        if sql_type is None:
            return field_type
        return sql_type

    def __str__(self):
        return f"{self.table_name}.{self.field_name}"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            operator=SqliteDialect.equals,
            tbl_name=self.table_name,
            driver=self.driver,
            first=self,
            second=other,
        )

    def __ne__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            operator=SqliteDialect.neq,
            driver=self.driver,
            first=self,
            second=other,
            tbl_name=self.table_name,
        )

    def __lt__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            operator=SqliteDialect.lt,
            driver=self.driver,
            first=self,
            second=other,
            tbl_name=self.table_name,
        )

    def __le__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            operator=SqliteDialect.le,
            driver=self.driver,
            first=self,
            second=other,
            tbl_name=self.table_name,
        )

    def __gt__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            operator=SqliteDialect.gt,
            driver=self.driver,
            first=self,
            second=other,
            tbl_name=self.table_name,
        )

    def __ge__(self, other):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            operator=SqliteDialect.ge,
            driver=self.driver,
            first=self,
            second=other,
            tbl_name=self.table_name,
        )


class Table:
    def __init__(
        self,
        db=None,
        conn=None,
        cursor=None,
        table_name: str = "",
        fields: list = None,
        dialect=None,
        driver=None,
    ):
        self.db = db
        self.conn = conn
        self.cursor = cursor
        self.table_name = table_name
        self.fields: dict = {field.field_name: field for field in fields}
        self.dialect = dialect
        self.driver = driver

    def insert(self, **values):
        return Query(
            db=self.db,
            conn=self.conn,
            cursor=self.cursor,
            dialect=SqliteDialect,
            driver=self.driver,
            tbl_name=self.table_name,
        ).insert(**values)

    def select(self, *fields, distinct=False, orderby=None):
        return Query(
            db=self.db,
            cursor=self.cursor,
            dialect=self.dialect,
            driver=self.driver,
            tbl_name=self.table_name,
        ).select(*fields, distinct=distinct, orderby=orderby)

    def update(self, **kwargs):
        return Query(
            db=self.db,
            cursor=self.cursor,
            dialect=self.dialect,
            driver=self.driver,
            tbl_name=self.table_name,
        ).update(**kwargs)

    def __getattribute__(self, item):
        try:
            return super(Table, self).__getattribute__(item)
        except Exception as e:
            if item in self.fields.keys():
                return self.fields[item]

            raise e
